//---------------------------------------------------------------------------
// smart_ptr.h
// Smart Pointer declaration and definition
// Developed by E. Nazarov
// http://nevsoft.narod.ru/
//---------------------------------------------------------------------------
#ifndef smart_ptrH
#define smart_ptrH
#include <utility>
#include <list>
//---------------------------------------------------------------------------
template<class T> class smart_ptr
{
private:
		typedef typename std::pair<int,T*>    pair_type;
		typedef typename std::list<pair_type> ptr_list;
        static ptr_list ptrs;
		             T *ptr;
public:
         smart_ptr(void);
         smart_ptr(T*);
         smart_ptr(smart_ptr<T>&);
        ~smart_ptr(void);
public:
        void reset(T *_ptr = NULL) throw();
        T*   get(void)             throw();
        T*   release(void)         throw();
        int  counter(void)  const  throw();
public:
     T&            operator*  () const throw();
     T*            operator-> () const throw();
     smart_ptr<T>  operator=  (smart_ptr<T>&      )       throw();
     bool          operator!= (smart_ptr<T> const&) const throw();
     bool          operator== (smart_ptr<T> const&) const throw();
};

template<class T> typename smart_ptr<T>::ptr_list smart_ptr<T>::ptrs = smart_ptr<T>::ptr_list();

template<class T>
smart_ptr<T>::smart_ptr(void)
{ ptr = NULL; }

template<class T>
smart_ptr<T>::smart_ptr(T* _ptr)
{
	if((ptr=_ptr)==NULL) { return; }
	for(ptr_list::iterator i=ptrs.begin(); i!=ptrs.end(); ++i)
		if((*i).second==_ptr) { ++((*i).first); return; }
	pair_type new_pair;
	new_pair.first  = 1;
	new_pair.second = _ptr;
	ptrs.push_back(new_pair);
}

template<class T>
smart_ptr<T>::smart_ptr(smart_ptr<T> &from)
{ ptr = from.ptr; from.ptr = NULL; }

template<class T>
smart_ptr<T>::~smart_ptr(void)
{
	if(ptr!=NULL) for(ptr_list::iterator i=ptrs.begin(); i!=ptrs.end(); ++i)
		if(i->second==ptr) 
			if(--(i->first)==0) { ptrs.erase(i); delete ptr; break; }
}

template<class T>
void smart_ptr<T>::reset(T *_ptr) throw()
{
	if(ptr==_ptr) return;
	for(ptr_list::iterator i=ptrs.begin(); i!=ptrs.end(); ++i)
		if(i->second==ptr) 
			if(--(i->first)==0) { delete ptr; ptrs.erase(i); break; }
	if((ptr=_ptr)==NULL) return;
	for(ptr_list::iterator i=ptrs.begin(); i!=ptrs.end(); ++i)
	if(i->second==_ptr) { ptr = _ptr; ++(i->first); return; }
	pair_type new_pair;
	new_pair.first  = 1;
	new_pair.second = ptr = _ptr;
	ptrs.push_back(new_pair);
}

template<class T>
T* smart_ptr<T>::get(void) throw()
{ return ptr; }

template<class T>
T* smart_ptr<T>::release(void) throw()
{
	for(ptr_list::iterator i=ptrs.begin(); i!=ptrs.end(); ++i)
		if(i->second==ptr) 
			if(--(i->first)==0) { ptrs.erase(i); break; }
	T* tmp = ptr;
	ptr = NULL;
	return tmp;
}

template<class T>
int smart_ptr<T>::counter(void) const throw()
{
	for(ptr_list::iterator i=ptrs.begin(); i!=ptrs.end(); ++i)
		if(i->second==ptr) 
			return i->first;
	return 0;
}

template<class T>
T& smart_ptr<T>::operator* () const throw()
{ return *ptr; }

template<class T>
T* smart_ptr<T>::operator-> () const throw()
{ return ptr; }

template<class T>
smart_ptr<T> smart_ptr<T>::operator=(smart_ptr<T> &_ptr) throw()
{
	reset(_ptr.ptr);
	_ptr.ptr = NULL;
	return *this;
}

template<class T>
bool smart_ptr<T>::operator==(smart_ptr<T> const &_ptr) const throw()
{ return ptr==_ptr.ptr; }

template<class T>
bool smart_ptr<T>::operator!=(smart_ptr<T> const &_ptr) const throw()
{ return ptr!=_ptr.ptr; }

#endif
 